﻿/*******************************************************************************
* Copyright (C) 2014 - 2015 Intel Corp. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Intel Corp. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Intel Corp. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Management;
using System.IO;

namespace FwUpdateAPI
{
    /// <summary>
    /// This class is a base class for the interfacing with both controllers and devices.
    ///
    /// This, with the derived classes, is the main interface of this API module for applications
    /// to use.
    /// The main role of this class is to wrap the interface to WMI but it includes also the
    /// interface for validating image compatibility with the current controller.
    /// </summary>
    public abstract class SdkTbtBase
    {
        /// <summary>
        /// C-tor. For use by derived classes.
        /// </summary>
        /// <param name="mo">WMI object to use for interfacing with the controller/device</param>
        protected SdkTbtBase(ManagementObject mo)
        {
            Mo = mo;
        }

        /// <summary>
        /// Abstract base property to get the proper message when a controller needs powering down.
        /// </summary>
        /// <returns>Child classes return appropriate message string</returns>
        public abstract string GetNeedPowerDownMessage { get; }

        /// <summary>
        /// Validates FW image file compatibility with the current controller
        /// </summary>
        /// <param name="path">Path to FW image file to check</param>
        /// <remarks>
        /// When controller is in safe-mode, it's impossible to read its FW to gather needed info
        /// for comparison, so the function is no-op! User has to check if the controller is in
        /// safe-mode to know if validation had any effect.
        /// </remarks>
        /// <exception>Throws exception in case of incompatibility error</exception>
        public abstract void ValidateImage(string path);

        /// <summary>
        /// WMI object to use for interfacing with the controller/device
        /// </summary>
        protected ManagementObject Mo { get; private set; }

        /// <summary>
        /// Starts the NVM update process
        /// </summary>
        /// <param name="bufferSize">Size of the array passed in buffer argument</param>
        /// <param name="buffer">Buffer with the firmware image binary to use for the update</param>
        /// <exception cref="TbtException">
        /// Throws an exception with proper message and the returned error code in case of an error
        /// </exception>
        public void UpdateFirmware(uint bufferSize, byte[] buffer)
        {
            const string methodName = "UpdateFirmware";
            var inparams = Mo.GetMethodParameters(methodName);
            inparams["bufferSize"] = bufferSize;
            inparams["buffer"] = buffer;
            var outParams = Mo.InvokeMethod(methodName, inparams, null);

            var result = (TbtStatus)outParams.GetPropertyValue("ReturnValue");
            if (result != TbtStatus.SUCCESS_RESPONSE_CODE)
            {
                throw new TbtException(result, Resources.FWUpdateFailedMessage);
            }
        }

        /// <summary>
        /// Reads the current NVM version from the controller/device
        /// </summary>
        /// <returns>NVM version as returned from the firmware</returns>
        /// <remarks>
        /// For printing the version, refer to Utilities.NvmVersionToString()
        /// </remarks>
        /// <exception cref="TbtException">
        /// Throws an exception with proper message and the returned error code in case of an error
        /// </exception>
        public virtual UInt32 GetCurrentNvmVersion()
        {
            var outParams = Mo.InvokeMethod("GetCurrentNvmVersion", null, null);
            var result = (TbtStatus)outParams.GetPropertyValue("ReturnValue");
            if (result != TbtStatus.SUCCESS_RESPONSE_CODE)
            {
                throw new TbtException(result, Resources.GetFirmwareVersionFailed);
            }
            return (UInt32)outParams.GetPropertyValue("nvmVersion");
        }

        /// <summary>
        /// Reads the current PD version from the controller/device
        /// </summary>
        /// <returns>PD version as returned from the firmware</returns>
        /// <exception cref="TbtException">
        /// Throws an exception with proper message and the returned error code in case of an error
        /// </exception>
        public virtual string GetCurrentPdVersion()
        {
            var outParams = Mo.InvokeMethod("GetCurrentPDVersion", null, null);
            var result = (TbtStatus)outParams.GetPropertyValue("ReturnValue");
            if (result != TbtStatus.SUCCESS_RESPONSE_CODE)
            {
                throw new TbtException(result, Resources.GetPdVersionFailed);
            }
            return (string)outParams.GetPropertyValue("pdVersion");
        }

        /// <summary>
        /// Reads data from NVM (FW) of the controller/device
        /// </summary>
        /// <param name="offset">Offset in NVM</param>
        /// <param name="length">Length to read (in bytes)</param>
        /// <returns>Returns byte array with the information read from firmware</returns>
        /// <exception>Throws on failure</exception>
        /// <remarks>
        /// The service supports only DW aligned offsets and reads 1 DW (4 bytes) for each read.
        /// This function gives simpler interface to the rest of the code with support to any offset
        /// and length as needed by doing the alignment internally, reading in loop the needed data
        /// and then extracting only the requested data for returning
        /// </remarks>
        public virtual byte[] ReadFirmware(UInt32 offset, UInt32 length)
        {
            const string methodName = "ReadFirmware";
            const int readSize = sizeof(UInt32);    // The size of each read from FW (DW, 4 bytes)
            const int lengthParameter = 1;          // The length parameter for ReadFirmware WMI call is always 1

            var inparams = Mo.GetMethodParameters(methodName);
            inparams["length"] = lengthParameter;

            var content = new List<byte>();         // Holds the data read from FW

            var reminder = offset % sizeof(UInt32); // Holds the needed change for aligning the offset to DW boundaries
            for (var alignedOffset = offset - reminder; alignedOffset < offset + length;
                 alignedOffset += readSize)
            {
                inparams["offset"] = alignedOffset;
                var outParams = Mo.InvokeMethod(methodName, inparams, null);
                var result = (TbtStatus)outParams.GetPropertyValue("ReturnValue");
                if (result != 0)
                {
                    throw new TbtException(result, Resources.ReadFirmwareFailed);
                }
                content.AddRange((byte[])outParams.GetPropertyValue("data"));
            }

            // Get only the requested data
            // Skip() ignores the possible added bytes at the start for offset alignment
            // Take() takes only the requested length, ignoring possible added bytes at the end due
            // to both the reading in 4-bytes group and the offset alignment
            return content.Skip((Int32)reminder).Take((Int32)length).ToArray();
        }

        /// <summary>Wrapper for SDK API to update from a file</summary>
        /// <param name="filename">Path to firmware image file</param>
        /// <exception>Throws an exception in case of an error</exception>
        public void UpdateFirmwareFromFile(string filename)
        {
            using (var fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
            {
                var reader = new BinaryReader(fs);
                var imageBuffer = reader.ReadBytes((int)fs.Length);
                UpdateFirmware((UInt32)imageBuffer.Length, imageBuffer);
            }
        }
    }
}
